from __future__ import print_function
import numpy as np
import tflearn
import tensorflow as tf
import random
import os
import sys
import csv
from sklearn.cross_validation import train_test_split
import experiments

def parse_csv(filename, feature_columns, label_column=None, dtype=np.float32):
	data, labels = [], []
	with open(filename, "r") as file:
		reader = csv.reader(file, delimiter=",")
		header = {v:k for k, v in enumerate(next(reader))}
		for row in reader:
			data.append([row[header[i]] for i in feature_columns])
			if label_column is not None:
				labels.append(row[header[label_column]])
	return np.array(data, dtype=dtype), tflearn.data_utils.to_categorical(labels, 2)

def split_data(data, labels):
	X_rest, X_test, y_rest, y_test = train_test_split(data, labels,
		test_size=0.1, random_state=42)
	X_train, X_val, y_train, y_val = train_test_split(X_rest, y_rest,
		test_size=0.1, random_state=42)
	
	return X_train, X_val, X_test, y_train, y_val, y_test

def get_checkpoint(dir):
	checkpoint_file = dir+"checkpoint"
	with open(checkpoint_file, "r") as file:
		checkpoint = file.readlines()[0].split("\"")[1]
		if not os.path.isabs(checkpoint):
			return dir+checkpoint

def train(name, dir, input, net, optimizer, batch_size, epochs):
	# Daten einlesen und aufteilen
	data, labels = parse_csv(*input)
	X_train, X_val, X_test, y_train, y_val, y_test = split_data(data, labels)
	
	# Preprocessing: zero-centern und normalisieren
	X_train, mean = tflearn.data_utils.featurewise_zero_center(X_train)
	X_train, std = tflearn.data_utils.featurewise_std_normalization(X_train)
	X_val = tflearn.data_utils.featurewise_zero_center(X_val, mean)
	X_val = tflearn.data_utils.featurewise_std_normalization(X_val, std)
	np.save(dir+"mean", mean)
	np.save(dir+"std", std)
	
	# Training
	net = tflearn.regression(net, optimizer=optimizer)
	model = tflearn.DNN(net, best_checkpoint_path=dir)
	model.fit(X_train, y_train, validation_set=(X_val, y_val), show_metric=True,
		n_epoch=epochs, batch_size=batch_size, run_id=name)

def evaluate(name, dir, input, net, optimizer, batch_size, epochs):
	if not os.path.isfile(dir+"mean.npy"):
		sys.exit("Experiment %s wurde noch nicht trainiert" % name)
	
	# Daten einlesen und aufteilen
	data, labels = parse_csv(*input)
	X_train, X_val, X_test, y_train, y_val, y_test = split_data(data, labels)
	
	# Preprocessing: zero-centern und normalisieren
	mean, std = np.load(dir+"mean.npy"), np.load(dir+"std.npy")
	X_test = tflearn.data_utils.featurewise_zero_center(X_test, mean)
	X_test = tflearn.data_utils.featurewise_std_normalization(X_test, std)
	
	net = tflearn.regression(net, optimizer=optimizer)
	model = tflearn.DNN(net, best_checkpoint_path=dir)
	model.load(get_checkpoint(dir))
	
	print(model.evaluate(X_test, y_test))

def predict(name, dir, input, net, optimizer, batch_size, epochs):
	if not os.path.isfile(dir+"mean.npy"):
		sys.exit("Experiment %s wurde noch nicht trainiert" % name)
	
	if len(sys.argv) < 4:
		sys.exit("Usage: python dnn.py predict <experiment> <input-csv>")
	
	# Daten einlesen
	X, _ = parse_csv(sys.argv[3], input[1])
	
	# Preprocessing: zero-centern und normalisieren
	mean, std = np.load(dir+"mean.npy"), np.load(dir+"std.npy")
	X = tflearn.data_utils.featurewise_zero_center(X, mean)
	X = tflearn.data_utils.featurewise_std_normalization(X, std)
	
	net = tflearn.regression(net, optimizer=optimizer)
	model = tflearn.DNN(net, best_checkpoint_path=dir)
	model.load(get_checkpoint(dir))
	
	for line in model.predict(X):
		print(line)

USAGE = "Usage: python dnn.py (train|evaluate|predict) <experiment> [<input-csv>]"

if __name__ == "__main__":
	# fuer reproduzierbare Ergebnisse RNGs mit festem Seed initialisieren
	random.seed(42)
	np.random.seed(42)
	tflearn.init_graph(seed=42)
	tf.set_random_seed(42)
	
	if len(sys.argv) < 3:
		sys.exit(USAGE)
	
	experiment_name = sys.argv[2]
	try:
		experiment = getattr(experiments, experiment_name)()
	except AttributeError:
		sys.exit("Experiment nicht gefunden")
	
	checkpoint_dir = "./checkpoints/"+experiment_name+"/"
	if not os.path.isdir(checkpoint_dir):
		os.makedirs(checkpoint_dir)
	
	if sys.argv[1] == "train":
		train(experiment_name, checkpoint_dir, *experiment)
	elif sys.argv[1] == "evaluate":
		evaluate(experiment_name, checkpoint_dir, *experiment)
	elif sys.argv[1] == "predict":
		predict(experiment_name, checkpoint_dir, *experiment)
	else:
		sys.exit(USAGE)
